from pyglossary import core
from pyglossary.core import rootDir
import logging
from io import BytesIO
from io import StringIO

log = logging.getLogger("pyglossary")


class XdxfTransformer(object):
	_gram_color: str = "green"
	_example_padding: int = 10

	def __init__(self, encoding="utf-8"):
		self._encoding = encoding

	def tostring(self, elem: "lxml.etree.Element") -> str:
		from lxml import etree as ET
		return ET.tostring(
			elem,
			method="html",
			pretty_print=True,
		).decode("utf-8").strip()

	def hasPrevText(self, prev: "Union[None, str, lxml.etree.Element]"):
		if isinstance(prev, str):
			return True
		if prev is None:
			return False
		if prev.tag == "k":
			return False
		if prev.tag in (
			"dtrn", "def", "span", "co",
			"i", "b", "sub", "sup", "tt", "big", "small",
		):
			return True
		if prev.text:
			return True
		# print(prev)
		return False

	def writeString(
		self,
		hf: "lxml.etree.htmlfile",
		child: str,
		parent: "lxml.etree.Element",
		prev: "Union[None, str, lxml.etree.Element]",
	):
		from lxml import etree as ET
		hasPrev = self.hasPrevText(prev)
		trailNL = False
		if parent.tag in ("ar", "font"):
			if child.startswith("\n"):
				child = child.lstrip("\n")
				if hasPrev:
					hf.write(ET.Element("br"))
			elif child.endswith("\n"):
				child = child.rstrip("\n")
				trailNL = True
			if not hasPrev:
				child = child.lstrip()
		elif child.startswith("\n"):
			child = child.lstrip()
			hf.write(ET.Element("br"))

		for index, parag in enumerate(child.split("\n")):
			if index > 0:
				hf.write(ET.Element("br"))
			hf.write(parag)
		if trailNL:
			hf.write(ET.Element("br"))
		return

	def writeExample(
		self,
		hf: "lxml.etree.htmlfile",
		elem: "Union[str, lxml.etree.Element]",
	):
		with hf.element("div", **{
			"class": "example",
			"style": f"padding: {self._example_padding}px 0px;",
		}):
			for child in elem.xpath("child::node()"):
				if isinstance(child, str):
					continue
				if child.tag == "iref":
					with hf.element("div"):
						self.writeIRef(hf, child)
					continue
				if child.tag in ("ex_orig", "ex_tran"):
					with hf.element("div"):
						self.writeChildrenOf(hf, child)
					continue
				log.warning(f"unknown tag {child.tag} inside <ex>")

	def writeIRef(
		self,
		hf: "lxml.etree.htmlfile",
		child: "Union[str, lxml.etree.Element]",
	):
		iref_url = child.attrib.get("href", "")
		if child.text:
			with hf.element("a", **{
				"class": "iref",
				"href": child.attrib.get("href", child.text),
			}):
				hf.write(child.text)
		elif any(iref_url.endswith(ext) for ext in ("mp3", "wav", "aac", "ogg")):
			#  with hf.element("audio", src=iref_url):
			with hf.element("a", **{
				"class": "iref",
				"href": iref_url,
			}):
				hf.write("🔊")
			return
		elif iref_url:
			with hf.element("a", **{
				"class": "iref",
				"href": iref_url,
			}):
				hf.write("⎋")
		else:
			log.warning(f"iref with no text and no url: {self.tostring(child)}")

	def writeChild(
		self,
		hf: "lxml.etree.htmlfile",
		child: "Union[str, lxml.etree.Element]",
		parent: "lxml.etree.Element",
		prev: "Union[None, str, lxml.etree.Element]",
	):
		from lxml import etree as ET

		if isinstance(child, str):
			self.writeString(hf, child, parent, prev)
			return

		if child.tag == f"br":
			hf.write(ET.Element("br"))
			return

		if child.tag in ("i", "b", "sub", "sup", "tt", "big", "small"):
			with hf.element(child.tag):
				self.writeChildrenOf(hf, child)
				# if child.text is not None:
				# 	hf.write(child.text.strip("\n"))
			return

		if child.tag == "blockquote":
			with hf.element("div", **{"class": "m"}):
				self.writeChildrenOf(hf, child)
			return

		if child.tag == "tr":
			hf.write("[")
			self.writeChildrenOf(hf, child)
			hf.write("]")
			return

		if child.tag in ("k", "sr"):
			with hf.element("div", **{"class": child.tag}):
				self.writeChildrenOf(hf, child)
			return

		if child.tag == "ex":
			self.writeExample(hf, child)
			return

		if child.tag == "mrkd":
			if child.text:
				return
			with hf.element("span", **{"class": child.tag}):
				with hf.element("b"):
					hf.write(child.text)
			return

		if child.tag in ("pos", "abr"):
			with hf.element("span", **{"class": "abr"}):
				with hf.element("font", color="green"):
					with hf.element("i"):
						self.writeChildrenOf(hf, child)
			return

		if child.tag in ("dtrn", "co"):
			self.writeChildrenOf(hf, child)
			return

		if child.tag == "c":
			color = child.attrib.get("c", "green")
			with hf.element("font", color=color):
				self.writeChildrenOf(hf, child)
			return

		if child.tag == "kref":
			if not child.text:
				log.warning(f"kref with no text: {self.tostring(child)}")
				return
			with hf.element("a", **{
				"class": "kref",
				"href": f"bword://{child.attrib.get('k', child.text)}",
			}):
				hf.write(child.text)
			return

		if child.tag == "iref":
			self.writeIRef(hf, child)
			return

		if child.tag == "rref":
			if not child.text:
				log.warning(f"rref with no text: {self.tostring(child)}")
				return

		if child.tag == "def":
			self.writeChildrenOf(hf, child)
			return

		if child.tag == "deftext":
			self.writeChildrenOf(hf, child)
			return

		if child.tag == "span":
			with hf.element("span"):
				self.writeChildrenOf(hf, child)
			return

		if child.tag == "abbr_def":
			# _type = child.attrib.get("type", "")
			# {"": "", "grm": "grammatical", "stl": "stylistical",
			#  "knl": "area/field of knowledge", "aux": "subsidiary"
			#  "oth": "others"}[_type]
			self.writeChildrenOf(hf, child)
			return

		if child.tag == "gr":
			with hf.element("font", color=self._gram_color):
				hf.write(child.text)
			return

		if child.tag == "ex_orig":
			with hf.element("i"):
				hf.write(child.text)
			return

		if child.tag == "ex_transl" and prev.tag == "ex_orig":
			if child.text != prev.text:
				with hf.element("i"):
					hf.write(child.text)
			return

		if child.tag == "opt":
			if child.text:
				hf.write(" (")
				hf.write(child.text)
				hf.write(")")
			return

		if child.tag == "img":
			with hf.element("img", **child.attrib):
				pass
			return

		if child.tag == "abbr":
			with hf.element("i"):
				hf.write(f"{child.text}")
			return

		log.warning(f"unknown tag {child.tag}")
		self.writeChildrenOf(hf, child)

	def writeChildrenOf(
		self,
		hf: "lxml.etree.htmlfile",
		elem: "lxml.etree.Element",
	):
		prev = None
		for child in elem.xpath("child::node()"):
			self.writeChild(hf, child, elem, prev)
			prev = child

	def transform(self, article: "lxml.etree.Element") -> str:
		from lxml import etree as ET
		encoding = self._encoding
		f = BytesIO()
		with ET.htmlfile(f, encoding="utf-8") as hf:
			with hf.element("div", **{"class": "article"}):
				self.writeChildrenOf(hf, article)

		text = f.getvalue().decode("utf-8")
		text = text.replace("<br>", "<br/>")  # for compatibility
		return text

	def transformByInnerString(self, articleInnerStr: str) -> str:
		from lxml import etree as ET
		return self.transform(
			ET.fromstring(f"<ar>{articleInnerStr}</ar>")
		)



# -*- coding: utf-8 -*-
# glossary_utils.py
#
# Copyright © 2008-2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import os
from os.path import (
	split,
	isdir,
	isfile,
	splitext,
)
import subprocess
import logging
import gc

from .compression import (
	compressionOpenFunc,
	stdCompressions,
)
from .entry import Entry

log = logging.getLogger("pyglossary")


class EntryList(object):
	def __init__(self, glos):
		self._l = []
		self._glos = glos
		self._sortKey = None

	def append(self, entry):
		self._l.append(entry.getRaw(self._glos))

	def insert(self, pos, entry):
		self._l.insert(pos, entry.getRaw(self._glos))

	def clear(self):
		self._l.clear()

	def __len__(self):
		return len(self._l)

	def __iter__(self):
		glos = self._glos
		for index, rawEntry in enumerate(self._l):
			if index & 0x7f == 0:  # 0x3f, 0x7f, 0xff
				gc.collect()
			yield Entry.fromRaw(
				glos, rawEntry,
				defaultDefiFormat=glos._defaultDefiFormat,
			)

	def setSortKey(self, sortKey, sampleItem):
		self._sortKey = Entry.getRawEntrySortKey(self._glos, sortKey)

	def sort(self):
		if self._sortKey is None:
			raise ValueError("EntryList.sort: sortKey is not set")
		self._l.sort(key=self._sortKey)

	def close(self):
		pass


def _zipFileAdd(zf, filename):
	if isfile(filename):
		zf.write(filename)
		return
	if not isdir(filename):
		raise OSError(f"Not a file or directory: {filename}")
	for subFname in os.listdir(filename):
		_zipFileAdd(zf, join(filename, subFname))


def zipFileOrDir(glos: "GlossaryType", filename: str) -> "Optional[str]":
	import zipfile
	import shutil
	from .os_utils import indir

	zf = zipfile.ZipFile(f"{filename}.zip", mode="w")

	if isdir(filename):
		dirn, name = split(filename)
		with indir(filename):
			for subFname in os.listdir(filename):
				_zipFileAdd(zf, subFname)

		shutil.rmtree(filename)
		return

	dirn, name = split(filename)
	files = [name]

	if isdir(f"{filename}_res"):
		files.append(f"{name}_res")

	with indir(dirn):
		for fname in files:
			_zipFileAdd(zf, fname)


def compress(glos: "GlossaryType", filename: str, compression: str) -> str:
	"""
	filename is the existing file path
	supported compressions: "gz", "bz2", "lzma", "zip"
	"""
	import shutil
	log.info(f"Compressing {filename!r} with {compression!r}")

	compFilename = f"{filename}.{compression}"
	if compression in stdCompressions:
		with compressionOpenFunc(compression)(compFilename, mode="wb") as dest:
			with open(filename, mode="rb") as source:
				shutil.copyfileobj(source, dest)
		return compFilename

	if compression == "zip":
		try:
			os.remove(compFilename)
		except OSError:
			pass
		try:
			error = zipFileOrDir(glos, filename)
		except Exception as e:
			log.error(
				f"{e}\nFailed to compress file \"{filename}\""
			)
	else:
		raise ValueError(f"unexpected compression={compression!r}")

	if isfile(compFilename):
		return compFilename
	else:
		return filename


def uncompress(srcFilename: str, dstFilename: str, compression: str) -> None:
	"""
	filename is the existing file path
	supported compressions: "gz", "bz2", "lzma"
	"""
	import shutil
	log.info(f"Uncompressing {srcFilename!r} to {dstFilename!r}")

	if compression in stdCompressions:
		with compressionOpenFunc(compression)(srcFilename, mode="rb") as source:
			with open(dstFilename, mode="wb") as dest:
				shutil.copyfileobj(source, dest)
		return

	# TODO: if compression == "zip":
	raise ValueError(f"unexpected compression={compression!r}")


def splitFilenameExt(
	filename: str = "",
) -> "Tuple[str, str, str]":
	"""
	returns (filenameNoExt, ext, compression)
	"""
	compression = ""
	filenameNoExt, ext = splitext(filename)
	ext = ext.lower()

	if not ext and len(filenameNoExt) < 5:
		filenameNoExt, ext = "", filenameNoExt

	if not ext:
		return filename, filename, "", ""

	if ext[1:] in stdCompressions + ("zip", "dz"):
		compression = ext[1:]
		filename = filenameNoExt
		filenameNoExt, ext = splitext(filename)
		ext = ext.lower()

	return filenameNoExt, filename, ext, compression



# -*- coding: utf-8 -*-

from formats_common import *

enable = False
lname = "gettext_mo"
format = "GettextMo"
description = "Gettext Binary (mo)"
extensions = (".mo",)
extensionCreate = ".mo"
singleFile = True
kind = "binary"
wiki = "https://en.wikipedia.org/wiki/Gettext"
website = (
	"https://www.gnu.org/software/gettext",
	"gettext - GNU Project",
)
optionsProp = {}



# -*- coding: utf-8 -*-
#
# Copyright © 2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from .option import Option
from .flags import (
	YesNoAlwaysNever,
	DEFAULT_NO,
)
import logging
from collections import OrderedDict as odict

log = logging.getLogger("pyglossary")


class PluginProp(object):
	def __init__(
		self,
		module,
	) -> None:
		self._mod = module
		self._Reader = None
		self._ReaderLoaded = False
		self._Writer = None
		self._WriterLoaded = False

		if log.level >= logging.DEBUG:
			for name, opt in self.optionsProp.items():
				if name.lower() != name:
					suggestName = "".join([
						"_" + x.lower() if x.isupper()
						else x
						for x in name
					])
					log.debug(
						f"{self.name}: please rename option "
						f"{name} to {suggestName}"
					)
				if not opt.comment:
					log.debug(
						f"{self.name}: please add comment for option {name}"
					)

	@property
	def pluginModule(self):
		return self._mod

	@property
	def lname(self) -> str:
		return self._mod.lname

	@property
	def name(self) -> str:
		return self._mod.format

	@property
	def description(self) -> str:
		return self._mod.description

	@property
	def extensions(self) -> "Tuple[str, ...]":
		return self._mod.extensions

	@property
	def ext(self) -> str:
		extensions = self.extensions
		if extensions:
			return extensions[0]
		return ""

	@property
	def extensionCreate(self) -> str:
		return self._mod.extensionCreate

	@property
	def singleFile(self) -> bool:
		return self._mod.singleFile

	@property
	def optionsProp(self) -> "Dict[str, Option]":
		return getattr(self._mod, "optionsProp", {})

	@property
	def sortOnWrite(self) -> YesNoAlwaysNever:
		return getattr(self._mod, "sortOnWrite", DEFAULT_NO)

	@property
	def sortKeyName(self) -> "Optional[Callable]":
		return getattr(self._mod, "sortKeyName", None)

	@property
	def path(self) -> "pathlib.Path":
		from pathlib import Path
		return Path(self._mod.__file__)

	def _loadReaderClass(self) -> "Optional[Any]":
		cls = getattr(self._mod, "Reader", None)
		if cls is None:
			return None
		for attr in (
			"__init__",
			"open",
			"close",
			"__len__",
			"__iter__",
		):
			if not hasattr(cls, attr):
				log.error(
					f"Invalid Reader class in {self.name!r} plugin"
					f", no {attr!r} method"
				)
				self._mod.Reader = None
				return None

		if hasattr(cls, "depends"):
			if not isinstance(cls.depends, dict):
				log.error(
					f"invalid depends={cls.depends}"
					f" in {self.name!r}.Reader class"
				)
		else:
			cls.depends = {}

		return cls

	@property
	def readerClass(self) -> "Optional[Any]":
		if self._ReaderLoaded:
			return self._Reader
		cls = self._loadReaderClass()
		self._Reader = cls
		self._ReaderLoaded = True
		return cls

	def _loadWriterClass(self) -> "Optional[Any]":
		cls = getattr(self._mod, "Writer", None)
		if cls is None:
			return None
		for attr in (
			"__init__",
			"open",
			"write",
			"finish",
		):
			if not hasattr(cls, attr):
				log.error(
					f"Invalid Writer class in {self.name!r} plugin"
					f", no {attr!r} method"
				)
				self._mod.Writer = None
				return None

		if hasattr(cls, "depends"):
			if not isinstance(cls.depends, dict):
				log.error(
					f"invalid depends={cls.depends}"
					f" in {self.name!r}.Writer class"
				)
		else:
			cls.depends = {}

		return cls

	@property
	def writerClass(self) -> "Optional[Any]":
		if self._WriterLoaded:
			return self._Writer
		cls = self._loadWriterClass()
		self._Writer = cls
		self._WriterLoaded = True
		return cls

	@property
	def canRead(self) -> bool:
		return self.readerClass is not None

	@property
	def canWrite(self) -> bool:
		return self.writerClass is not None

	def getReadOptions(self):
		return self.getOptionsFromClass(self.readerClass)

	def getWriteOptions(self):
		return self.getOptionsFromClass(self.writerClass)

	def getOptionAttrNamesFromClass(self, rwclass):
		nameList = []

		for cls in rwclass.__bases__ + (rwclass,):
			for _name in cls.__dict__:
				if not _name.startswith("_") or _name.startswith("__"):
					# and _name not in ("_open",)
					continue
				nameList.append(_name)

		# rwclass.__dict__ does not include attributes of parent/base class
		# and dir(rwclass) is sorted by attribute name alphabetically
		# using rwclass.__bases__ solves the problem

		return nameList

	def getOptionsFromClass(self, rwclass):
		optionsProp = self.optionsProp
		options = odict()
		if rwclass is None:
			return options

		for attrName in self.getOptionAttrNamesFromClass(rwclass):
			name = attrName[1:]
			default = getattr(rwclass, attrName)
			if name not in optionsProp:
				if not callable(default):
					log.warning(
						f"format={self.name}, attrName={attrName}, type={type(default)}"
					)
				continue
			prop = optionsProp[name]
			if prop.disabled:
				log.trace(f"skipping disabled option {name} in {self.name} plugin")
				continue
			if not prop.validate(default):
				log.warning(
					"invalid default value for option: "
					f"{name} = {default!r} in plugin {self.name}"
				)
			options[name] = default

		return options

	def getReadExtraOptions(self):
		return self.__class__.getExtraOptions(self.readerClass.open, self.name)

	def getWriteExtraOptions(self):
		return self.__class__.getExtraOptions(self.writerClass.write, self.name)

	@classmethod
	def getExtraOptions(cls, func, format):
		import inspect
		extraOptNames = []
		for name, param in inspect.signature(func).parameters.items():
			if name == "self":
				continue
			if str(param.default) != "<class 'inspect._empty'>":
				extraOptNames.append(name)
				continue
			if name not in ("filename", "dirname"):
				extraOptNames.append(name)
		if extraOptNames:
			log.warning(f"{format}: extraOptNames = {extraOptNames}")
		return extraOptNames



# -*- coding: utf-8 -*-
#
# Copyright © 2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from collections import namedtuple
from operator import itemgetter
import re


NamedSortKey = namedtuple("NamedSortKey", [
	"name",
	"normal",
	"sqlite",
	"desc",
])


"""
sortKeyType = Callable[
	[[List[str]],
	Any,
]

sqliteSortKeyType = List[Tuple[str, str, sortKeyType]]
"""


def _headword_normal(encoding: str, **options) -> "sortKeyType":
	def sortKey(words: "List[str]"):
		return words[0].encode(encoding, errors="replace")

	return sortKey


def _headword_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	def sortKey(words: "List[str]"):
		return words[0].encode(encoding, errors="replace")

	return [
		(
			"headword",
			"TEXT" if encoding == "utf-8" else "BLOB",
			sortKey,
		),
	]


def _headword_lower_normal(encoding: str, **options) -> "sortKeyType":
	def sortKey(words: "List[str]"):
		return words[0].lower().encode(encoding, errors="replace")

	return sortKey


def _headword_lower_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	def sortKey(words: "List[str]"):
		return words[0].lower().encode(encoding, errors="replace")

	return [
		(
			"headword_lower",
			"TEXT" if encoding == "utf-8" else "BLOB",
			sortKey,
		),
	]


def _headword_bytes_lower_normal(encoding: str, **options) -> "sortKeyType":
	def sortKey(words: "List[str]"):
		return words[0].encode(encoding, errors="replace").lower()

	return sortKey


def _headword_bytes_lower_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	def sortKey(words: "List[str]"):
		return words[0].encode(encoding, errors="replace").lower()

	return [
		(
			"headword_blower",
			"TEXT" if encoding == "utf-8" else "BLOB",
			sortKey,
		),
	]


def _stardict_normal(encoding: str, **options) -> "sortKeyType":
	def sortKey(words: "List[str]"):
		b_word = words[0].encode(encoding, errors="replace")
		return (b_word.lower(), b_word)

	return sortKey


def _stardict_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	def headword_lower(words: "List[str]"):
		return words[0].encode(encoding, errors="replace").lower()

	def headword(words: "List[str]"):
		return words[0].encode(encoding, errors="replace")

	_type = "TEXT" if encoding == "utf-8" else "BLOB"
	return [
		(
			"headword_lower",
			_type,
			headword_lower,
		),
		(
			"headword",
			_type,
			headword,
		),
	]


def _ebook_normal(encoding: str, **options) -> "sortKeyType":
	length = options.get("group_by_prefix_length", 2)

	def sortKey(words: "List[str]"):
		word = words[0]
		if not word:
			return "", ""
		prefix = word[:length].lower()
		if prefix[0] < "a":
			return "SPECIAL", word
		return prefix, word

	return sortKey


def _ebook_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	length = options.get("group_by_prefix_length", 2)

	def getPrefix(words: "List[str]"):
		word = words[0]
		if not word:
			return ""
		prefix = word[:length].lower()
		if prefix[0] < "a":
			return "SPECIAL"
		return prefix

	def headword(words: "List[str]"):
		return words[0].encode(encoding, errors="replace")

	_type = "TEXT" if encoding == "utf-8" else "BLOB"
	return [
		(
			"prefix",
			_type,
			getPrefix,
		),
		(
			"headword",
			_type,
			headword,
		),
	]


def _ebook_length3_normal(encoding: str, **options) -> "sortKeyType":
	return _ebook_normal(
		encoding,
		group_by_prefix_length=3,
	)


def _ebook_length3_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	return _ebook_sqlite(
		encoding,
		group_by_prefix_length=3,
	)


_dicformids_re_punc = re.compile(
	r"[!\"$§$%&/()=?´`\\{}\[\]^°+*~#'-_.:,;<>@]*",
	# FIXME: |
)


def _dicformids_normal(encoding: str, **options) -> "sortKeyType":
	re_punc = _dicformids_re_punc
	re_spaces = re.compile(" +")
	re_tabs = re.compile("\t+")

	def sortKey(words: "List[str]") -> "Any":
		word = words[0]
		word = word.strip()
		# looks like we need to remove tabs, because app gives error
		# but based on the java code, all punctuations should be removed
		# as well, including '|' which is used to separate alternate words
		# FIXME
		# word = word.replace("|", " ")
		word = re_punc.sub("", word)
		word = re_spaces.sub(" ", word)
		word = re_tabs.sub(" ", word)
		word = word.lower()
		return word

	return sortKey


def _dicformids_sqlite(encoding: str, **options) -> "sqliteSortKeyType":
	return [
		(
			"headword_norm",
			"TEXT",
			_dicformids_normal(encoding, **options),
		),
	]


namedSortKeyList = [
	NamedSortKey(
		name="headword",
		normal=_headword_normal,
		sqlite=_headword_sqlite,
		desc="Headword",
	),
	NamedSortKey(
		name="headword_lower",
		normal=_headword_lower_normal,
		sqlite=_headword_lower_sqlite,
		desc="Lowercase Headword",
	),
	NamedSortKey(
		name="headword_bytes_lower",
		normal=_headword_bytes_lower_normal,
		sqlite=_headword_bytes_lower_sqlite,
		desc="ASCII-Lowercase Headword",
	),
	NamedSortKey(
		name="stardict",
		normal=_stardict_normal,
		sqlite=_stardict_sqlite,
		desc="StarDict",
	),
	NamedSortKey(
		name="ebook",
		normal=_ebook_normal,
		sqlite=_ebook_sqlite,
		desc="E-Book (prefix length: 2)",
	),
	NamedSortKey(
		name="ebook_length3",
		normal=_ebook_length3_normal,
		sqlite=_ebook_length3_sqlite,
		desc="E-Book (prefix length: 3)",
	),
	NamedSortKey(
		name="dicformids",
		normal=_dicformids_normal,
		sqlite=_dicformids_sqlite,
		desc="DictionaryForMIDs",
	),
]

namedSortKeyByName = {
	item.name: item for item in namedSortKeyList
}

"""
https://en.wikipedia.org/wiki/UTF-8#Comparison_with_other_encodings

Sorting order: The chosen values of the leading bytes means that a list
of UTF-8 strings can be sorted in code point order by sorting the
corresponding byte sequences.
"""



class StrWithDesc(str):
	def __new__(cls, name: str, desc: str):
		s = str.__new__(cls, name)
		s.desc = desc
		return s


ALWAYS = StrWithDesc("always", "Always")
DEFAULT_YES = StrWithDesc("default_yes", "Yes (by default)")
DEFAULT_NO = StrWithDesc("default_no", "No (by default)")
NEVER = StrWithDesc("never", "Never")

# typing.Literal is added in Python 3.8
YesNoAlwaysNever = """Union[
	"Literal[ALWAYS]",
	"Literal[DEFAULT_YES]",
	"Literal[DEFAULT_NO]",
	"Literal[NEVER]",
]"""



# coding=utf-8
# based on https://github.com/zkoch/CEDICT_Parser

TONES = {
	"a1": "ā", "a2": "á", "a3": "ǎ", "a4": "à",
	"e1": "ē", "e2": "é", "e3": "ě", "e4": "è",
	"i1": "ī", "i2": "í", "i3": "ǐ", "i4": "ì",
	"o1": "ō", "o2": "ó", "o3": "ǒ", "o4": "ò",
	"u1": "ū", "u2": "ú", "u3": "ǔ", "u4": "ù",
	"v1": "ǖ", "v2": "ǘ", "v3": "ǚ", "v4": "ǜ",
}
# using v for the umlauted u

VOWELS = ("a", "e", "o", "iu", "ui", "i", "u", "v")


def convert(word):
	tone = word[-1]
	pinyin = word[0:-1].lower()
	result = pinyin

	if tone == "5":
		return pinyin, tone
	elif tone not in ("1", "2", "3", "4"):
		return word, ""

	for vowel in VOWELS:
		if vowel in pinyin:
			vowel1 = vowel[-1]
			result = pinyin.replace(vowel1, TONES[vowel1 + tone])
			break

	return result, tone



# -*- coding: utf-8 -*-
from formats_common import *

enable = False
lname = "babylon_bdc"
format = "BabylonBdc"
description = "Babylon (bdc)"
extensions = (".bdc",)
extensionCreate = ""
singleFile = True
kind = "binary"
wiki = ""
website = None
optionsProp = {}



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from pickle import dumps, loads
import os
from os.path import isfile
import gc
from .entry import Entry

import logging
log = logging.getLogger("pyglossary")

PICKLE_PROTOCOL = 4

# Pickle protocol 4 performed better than protocol 5 on Python 3.9.2
# Slightly lower running time, lower memory usage, and same .db file size

# Pickle protocol 5		added in Python 3.8		PEP 574
# Pickle protocol 4		added in Python 3.4		PEP 3154
# Pickle Protocol 3		added in Python 3.0

# https://docs.python.org/3/library/pickle.html


class SqEntryList(list):
	def __init__(
		self,
		glos,
		filename: str,
		sortColumns: "List[Tuple[str, str, str]]",
		create: bool = True,
		persist: bool = False,
	):
		"""
			sortColumns[i] == (name, type, valueFunc)

			persist: do not delete the file when variable is deleted
		"""
		from sqlite3 import connect

		self._glos = glos
		self._filename = filename
		self._persist = persist
		self._con = connect(filename)
		self._cur = self._con.cursor()

		if not filename:
			raise ValueError(f"invalid filename={filename!r}")

		self._sortColumns = sortColumns
		self._columnNames = ",".join([
			col[0] for col in sortColumns
		])
		self._orderBy = "rowid"
		self._sorted = False
		self._reverse = False
		self._len = 0
		if create:
			colDefs = ",".join([
				f"{col[0]} {col[1]}"
				for col in sortColumns
			] + ["pickle BLOB"])
			self._con.execute(
				f"CREATE TABLE data ({colDefs})"
			)
		else:
			self._parseExistingIndex()

	def __len__(self):
		return self._len

	def append(self, entry):
		rawEntry = entry.getRaw(self._glos)
		self._len += 1
		colCount = len(self._sortColumns)
		try:
			values = [
				col[2](entry.l_word) for col in self._sortColumns
			]
		except Exception:
			log.critical(f"error in _sortColumns funcs for rawEntry = {rawEntry!r}")
			raise
		try:
			pickleEntry = dumps(rawEntry, protocol=PICKLE_PROTOCOL)
		except Exception:
			log.critical(f"error in pickle.dumps for rawEntry = {rawEntry!r}")
			raise
		self._cur.execute(
			f"insert into data({self._columnNames}, pickle)"
			f" values (?{', ?' * colCount})",
			values + [pickleEntry],
		)
		if self._len % 1000 == 0:
			self._con.commit()

	def __iadd__(self, other):
		for item in other:
			self.append(item)
		return self

	def setSortKey(self, sortKey, sampleItem):
		log.warning(f"SqList: ignoring sortKey {sortKey}")
		# FIXME
		# sample = sortKey(sampleItem)

	def sort(self, reverse=False):
		if self._sorted:
			raise NotImplementedError("can not sort more than once")

		self._reverse = reverse
		self._sorted = True
		sortColumnNames = self._columnNames
		self._orderBy = sortColumnNames
		if reverse:
			self._orderBy = ",".join([
				f"{col[0]} DESC" for col in self._sortColumns
			])
		self._con.commit()
		self._con.execute(
			f"CREATE INDEX sortkey ON data({sortColumnNames});"
		)
		self._con.commit()

	def _parseExistingIndex(self) -> bool:
		self._cur.execute("select sql FROM sqlite_master WHERE name='sortkey'")
		row = self._cur.fetchone()
		if row is None:
			return False
		sql = row[0]
		# sql == "CREATE INDEX sortkey ON data(wordlower,word)"
		i = sql.find("(")
		if i < 0:
			log.error(f"error parsing index sql={sql!r}")
			return False
		j = sql.find(")", i)
		if j < 0:
			log.error(f"error parsing index sql={sql!r}")
			return False
		columnNames = sql[i + 1:j]
		self._sorted = True
		self._orderBy = columnNames
		return True

	def deleteAll(self):
		if self._con is None:
			return
		self._con.execute(
			f"DELETE FROM data;"
		)
		self._con.commit()
		self._len = 0

	def clear(self):
		self.close()

	def close(self):
		if self._con is None:
			return
		self._con.commit()
		self._cur.close()
		self._con.close()
		self._con = None
		self._cur = None

	def __del__(self):
		try:
			self.close()
			if not self._persist and isfile(self._filename):
				os.remove(self._filename)
		except AttributeError as e:
			log.error(str(e))

	def __iter__(self):
		glos = self._glos
		query = f"SELECT pickle FROM data ORDER BY {self._orderBy}"
		self._cur.execute(query)
		for index, row in enumerate(self._cur):
			if index & 0x7f == 0:  # 0x3f, 0x7f, 0xff
				gc.collect()
			yield Entry.fromRaw(
				glos, loads(row[0]),
				defaultDefiFormat=glos._defaultDefiFormat,
			)



# -*- coding: utf-8 -*-

from .entry_base import BaseEntry
from .entry import Entry, DataEntry
from .langs import Lang


class GlossaryType(object):
	"""
	an abstract type class for Glossary class in plugins. it only
	contains methods and properties that might be used in plugins
	"""

	def setDefaultDefiFormat(self, defiFormat: str) -> None:
		raise NotImplementedError

	def getDefaultDefiFormat(self) -> str:
		raise NotImplementedError

	def collectDefiFormat(
		self,
		maxCount: int,
	) -> "Optional[Dict[str, float]]":
		raise NotImplementedError

	def iterInfo(self) -> "Iterator[Tuple[str, str]]":
		raise NotImplementedError

	def getInfo(self, key: str) -> str:
		raise NotImplementedError

	def setInfo(self, key: str, value: str) -> None:
		raise NotImplementedError

	def getExtraInfos(self, excludeKeys: "List[str]") -> "OrderedDict":
		raise NotImplementedError

	@property
	def author(self) -> str:
		raise NotImplementedError

	@property
	def alts(self) -> bool:
		raise NotImplementedError

	@property
	def filename(self):
		raise NotImplementedError

	@property
	def sourceLang(self) -> "Optional[Lang]":
		raise NotImplementedError

	@property
	def targetLang(self) -> "Optional[Lang]":
		raise NotImplementedError

	@property
	def sourceLangName(self) -> str:
		raise NotImplementedError

	@sourceLangName.setter
	def sourceLangName(self, langName: str) -> None:
		raise NotImplementedError

	@property
	def targetLangName(self) -> str:
		raise NotImplementedError

	@targetLangName.setter
	def targetLangName(self, langName: str) -> None:
		raise NotImplementedError

	def titleElement(
		self,
		hf: "lxml.etree.htmlfile",
		sample: str = "",
	) -> "lxml.etree._FileWriterElement":
		raise NotImplementedError

	def wordTitleStr(
		self,
		word: str,
		sample: str = "",
		_class: str = "",
	) -> str:
		raise NotImplementedError

	def getConfig(self, name: str, default: "Optional[str]") -> "Optional[str]":
		raise NotImplementedError

	def addEntryObj(self, entry: "Entry") -> None:
		raise NotImplementedError

	def newEntry(self, word: str, defi: str, defiFormat: str = "") -> "Entry":
		raise NotImplementedError

	def newDataEntry(self, fname: str, data: bytes) -> DataEntry:
		raise NotImplementedError
